<?php

include_once __DIR__ .'/classes/Transaction.php';

class FirstDataPayment extends PaymentModule
{

	const GATEWAY_ADDR = 'https://vpos.polcard.com.pl/vpos/ecom/service.htm';
	const REST_ADDR = 'https://vpos.polcard.com.pl/vpos/ecommerce/rest';

	protected		$_html			= '';
	protected		$_errors		= array();
	
	
	// Config fields
	protected		$getewayAddr	= null;
	protected		$posid			= null;
	protected		$sharedKey		= null;
	protected		$mid			= null;
	protected		$user			= null;
	protected		$pass			= null;
	protected		$testMode		= true;
	
	protected		$logFile		= null;
	
	// Params
	protected		$_languages		= null;
	
	public			$token			= null;
	public			$fields_list	= null;
	public			$avaliableCurrencies = array('PLN', 'EUR');
	
	
	
	public function __construct()
	{
		$this->name = 'firstdatapayment';
		$this->tab = 'payments_gateways';
		$this->version = '1.11';
		$this->author = 'First Data Polcard';
		$this->controllers = array('payment', 'validation');
		$this->bootstrap = true;
        $this->is_eu_compatible = true;
		
		parent::__construct();

		# get params
		$config = Configuration::getMultiple(array(
			'FIRSTDATA_POSID',
			'FIRSTDATA_SHAREDKEY',
			'FIRSTDATA_MID',
			'FIRSTDATA_USER',
			'FIRSTDATA_PASS',
			'FIRSTDATA_TESTMODE',
			'FIRSTDATA_LOGFILE',
		));
		$this->posid = isset($config['FIRSTDATA_POSID']) ? $config['FIRSTDATA_POSID'] : null;
		$this->sharedKey = isset($config['FIRSTDATA_SHAREDKEY']) ? $config['FIRSTDATA_SHAREDKEY'] : null;
		$this->mid = isset($config['FIRSTDATA_MID']) ? $config['FIRSTDATA_MID'] : null;
		$this->user = isset($config['FIRSTDATA_USER']) ? $config['FIRSTDATA_USER'] : null;
		$this->pass = isset($config['FIRSTDATA_PASS']) ? $config['FIRSTDATA_PASS'] : null;
		$this->testMode = isset($config['FIRSTDATA_TESTMODE']) ? $config['FIRSTDATA_TESTMODE'] : false;
		$this->logFile = _PS_ROOT_DIR_ .'/log/firstdata_' . @$config['FIRSTDATA_LOGFILE'] .'.txt';

		$this->displayName = $this->l('FirstData payment geteway');
		
		if(empty($this->posid) || empty($this->sharedKey) || empty($this->mid) || empty($this->user) || empty($this->pass))
			$this->warning = $this->l('Account details must be configured before using this module.');
	}
	
	public function install()
	{
		$this->_languages = Language::getLanguages(false);
	
		# std
		if(	!parent::install() 
			|| !$this->registerHook('adminOrder')
			|| !$this->registerHook('displayPayment')
			|| !$this->registerHook('paymentOptions')
			|| !$this->registerHook('paymentReturn')
			)
			return false;
		
		# add states
		$state1 = new OrderState();
		$state1->module_name = $this->name;
		$state1->name = array();
		foreach($this->_languages as $lang)
			$state1->name[(int)$lang['id_lang']] = $this->l('Payment by FirstData - select method');
		$state1->color = '#4169E1';
		$state1->add();
		$state2 = new OrderState();
		$state2->module_name = $this->name;
		$state2->name = array();
		foreach($this->_languages as $lang)
			$state2->name[(int)$lang['id_lang']] = $this->l('Payment by FirstData - pending');
			$state2->color = '#4169E1';
			$state2->add();
		
		# copy state icon
		$source = dirname(__FILE__).'/logo.gif';
		$destination = dirname(__FILE__).'/../../img/os/'.(int)$state1->id.'.gif';
		copy($source, $destination);
		$destination = dirname(__FILE__).'/../../img/os/'.(int)$state2->id.'.gif';
		copy($source, $destination);
		
		# save state id
		Configuration::updateValue('FIRSTDATA_ORDERSTATE', $state1->id);
		Configuration::updateValue('FIRSTDATA_ORDERSTATE_PENDING', $state2->id);
		
		# create log file
		$logFile = uniqid();
		Configuration::updateValue('FIRSTDATA_LOGFILE', $logFile);
		
		# create transaction table
		Db::getInstance()->query("
			CREATE TABLE IF NOT EXISTS `"._DB_PREFIX_."firstdata_transaction` (
			  `id_transaction` int(10) unsigned NOT NULL AUTO_INCREMENT,
			  `id_order` int(10) unsigned NOT NULL,
			  `status` int(10) unsigned NOT NULL,
			  `error` int(10) unsigned NOT NULL,
			  `msg` VARCHAR( 1024 ) NULL,
			  `fd_tid` varchar(50) NOT NULL,
			  `date_add` datetime NOT NULL,
			  `date_mod` datetime DEFAULT NULL,
			  `date_accept` datetime DEFAULT NULL,
			  `date_cancel` datetime DEFAULT NULL,
			  PRIMARY KEY (`id_transaction`)
			) ENGINE=InnoDB  DEFAULT CHARSET=utf8 AUTO_INCREMENT=1;
		");
			
		return true;
	}
	
	public function uninstall()
	{
		$id_state1 = (int) Configuration::get('FIRSTDATA_ORDERSTATE');
		$id_state2 = (int) Configuration::get('FIRSTDATA_ORDERSTATE_PENDING');
		
		if(	!Configuration::deleteByName('FIRSTDATA_ORDERSTATE')
			|| ($id_state2 && !Configuration::deleteByName('FIRSTDATA_ORDERSTATE_PENDING'))
			|| !Configuration::deleteByName('FIRSTDATA_POSID')
			|| !Configuration::deleteByName('FIRSTDATA_SHAREDKEY')
			|| !Configuration::deleteByName('FIRSTDATA_MID')
			|| !Configuration::deleteByName('FIRSTDATA_USER')
			|| !Configuration::deleteByName('FIRSTDATA_PASS')
			|| !Configuration::deleteByName('FIRSTDATA_TESTMODE')
			|| !parent::uninstall() )
			return false;
			
		# remove state
		$state1 = new OrderState($id_state1);
		$state1->delete();
		if($id_state2) {
			$state2 = new OrderState($id_state2);
			$state2->delete();
		}
		
		# drop table
		Db::getInstance()->query("DROP TABLE IF EXISTS `"._DB_PREFIX_."firstdata_transaction`");
		
		return true;
	}
	
	public function getSharedKey() {
		return $this->sharedKey;
	}

	public function getTransactionStatuses()
	{
		return array(
			20 => $this->l('AUTHORIZATION_READY'),
			21 => $this->l('AUTHORIZATION_FAILED'),
			30 => $this->l('AUTHORIZED'),
			35 => $this->l('CLEARING_ACCEPTED'),
			40 => $this->l('REJECTED'),
			50 => $this->l('CLEARING_READY'),
			51 => $this->l('CLEARING_FAILED'),
			99 => $this->l('CANCELED'),
		);
	}

	protected function _postprocess()
	{
		if(Tools::isSubmit('saveConfig'))
		{
			$geteway = trim(Tools::getValue('FIRSTDATA_GETEWAYADDR'));
			if(strpos($geteway, 'vpos-test.polcard.com.pl/vpos') !== false && strpos($geteway, 'ecom/service.htm') === false)
			{
				$geteway .= $geteway[strlen($geteway)-1] == '/'
					? 'ecom/service.htm'
					: '/ecom/service.htm';
			}
			
			Configuration::updateValue('FIRSTDATA_GETEWAYADDR', $geteway);
			
			$posId = trim(Tools::getValue('FIRSTDATA_POSID'));
			Configuration::updateValue('FIRSTDATA_POSID', $posId);
			$sharedKey = trim(Tools::getValue('FIRSTDATA_SHAREDKEY'));
			Configuration::updateValue('FIRSTDATA_SHAREDKEY', $sharedKey);
			
			$mid = trim(Tools::getValue('FIRSTDATA_MID'));
			Configuration::updateValue('FIRSTDATA_MID', $mid);
			$user = trim(Tools::getValue('FIRSTDATA_USER'));
			Configuration::updateValue('FIRSTDATA_USER', $user);
			$pass = trim(Tools::getValue('FIRSTDATA_PASS'));
			if($pass) Configuration::updateValue('FIRSTDATA_PASS', $pass);
			
			$testMode = trim(Tools::getValue('FIRSTDATA_TESTMODE'));
			Configuration::updateValue('FIRSTDATA_TESTMODE', $testMode);
			$this->_html .= $this->displayConfirmation($this->l('Settings updated'));
		}
		if(Tools::isSubmit('fdp_refresh'))
		{
			$data = FirstDataTransaction::getPendingList();
			$err = false;
			if($data) {
				foreach ($data as $r) {
					$resp = $this->getTrans($r['id_order'], substr($r['date_add'], 0, 10));
					if(is_string($resp)) {
						$err = $resp;
						break;
					}

					if(($resp->orderCode == $r['id_order']) && ($resp->paymentStatus != $r['status'])) {	// Update transaction data
						$order = new Order($r['id_order']);
						$this->updateTrans($order, $resp->paymentStatus, '');						
					}
				}
			}
			
			$data = FirstDataTransaction::getNotRegisteredList(30);
			if($data) {
				foreach ($data as $r) {
					$order = new Order($r['id_order']);
					$bro = $order->getBrother();
					$hasBro = false;
					foreach($bro as $b) {
						$id_trans = FirstDataTransaction::getIdByOId($b->id);
						if($id_trans) {
							$hasBro = true;
							break;
						}
					}
					if($hasBro) continue;

					$resp = $this->getTrans($r['id_order'], substr($r['date_add'], 0, 10));
					if(!is_string($resp)) $stat = $resp->paymentStatus;
					else $stat = 999;
					$this->updateTrans($order, $stat, '');
				}
			}
				
			if(!$err) $this->_html .= $this->displayConfirmation($this->l('Transaction list updated'));
			else $this->_html .= $this->displayError($this->l($err));
		}
	}
	
	function callApi($req) {
		if(!$this->user || !$this->pass) return $this->l('User/password not set');
		if(!extension_loaded ('curl')) return $this->l('CURL extension not loaded');
		$url = self::REST_ADDR . $req;
		$curl = curl_init ();
		curl_setopt ( $curl, CURLOPT_URL, $url );
		curl_setopt ( $curl, CURLOPT_RETURNTRANSFER, 1 );
		curl_setopt ( $curl, CURLOPT_FAILONERROR, false );
		curl_setopt ( $curl, CURLOPT_USERPWD, $this->mid . '.' . $this->user . ":" . $this->pass);
	
		$result = curl_exec ( $curl );
		if($result === false) {
			$stat = curl_error($curl);
			curl_close ( $curl );
			return $stat;
		}
		$info = curl_getinfo($curl);
		if($info['http_code'] >= 400) {
			$stat = null;
			if($result) {
				$json = json_decode($result);
				$stat = $json->fault;
			}
			if($stat) $resp = $stat;
			elseif($info['http_code'] == 401) $resp = $this->l('Auth error');
			else $resp = 'HTTP error ' . $info['http_code'];
			curl_close ( $curl );
			return $resp;
		}
		curl_close ( $curl );
		return json_decode($result);
	}
	
	/*
	 * Configuration panel
	 * Process and display
	 */
	public function getContent()
	{
		# vars
		$this->token = Tools::getAdminTokenLite('AdminModules');
		$this->_languages = Language::getLanguages(false);
	
	
		$this->_html = '';
		
		# process
		$this->_postprocess();
		if($this->_errors)
		{
			foreach($this->_errors as $err)
				$this->_html .= $this->displayError($err);
			$this->_html .= '<br />';
		}
		
		# display
		$this->_html .= $this->display(__FILE__, 'config.tpl');
		$this->_html .= $this->_displayConfigForm();
		$this->_html .= $this->_displayTransactionList();
		
		$this->_html .= '';

		return $this->_html;
	}

	/*
	 * Render configuration form
	 */
	public function _displayConfigForm()
	{
		$ps16 = strpos(_PS_VERSION_, '1.6') === 0;
		$testInput = $ps16
			?	array(
					'type' => 'switch',
					'label' => $this->l('Test mode'),
					'name' => 'FIRSTDATA_TESTMODE',
					'is_bool' => true,
					'values' => array(
						array('id' => 'testmode_on', 'value' => 1, 'label' => $this->l('Enabled')),
						array('id' => 'testmode_off', 'value' => 0, 'label' => $this->l('Disabled')),
					),
				)
			:	array(
					'type' => 'select',
					'label' => $this->l('Test mode'),
					'name' => 'FIRSTDATA_TESTMODE',
					'options' => array(
						'query' => array(
							array('id' => 1, 'name' => $this->l('Enabled')),
							array('id' => 0, 'name' => $this->l('Disabled')),
						),
						'id' => 'id',
						'name' => 'name',
					),
				);
	
		$fields_form = array(
			'form' => array(
				'legend' => array(
					'title' => $this->l('Account details'),
					'icon' => 'icon-cogs'
				),
				'input' => array(
					array(
						'type' => 'text',
						'label' => $this->l('Shop Id'),
						'name' => 'FIRSTDATA_POSID',
					),
					array(
						'type' => 'text',
						'label' => $this->l('Shared Key'),
						'name' => 'FIRSTDATA_SHAREDKEY',
					),
					array(
						'type' => 'text',
						'label' => $this->l('MID'),
						'name' => 'FIRSTDATA_MID',
					),
					array(
						'type' => 'text',
						'label' => $this->l('User'),
						'name' => 'FIRSTDATA_USER',
					),
					array(
						'type' => 'password',
						'label' => $this->l('Password'),
						'name' => 'FIRSTDATA_PASS',
					),
					$testInput,
				),
				'submit' => array(
					'title' => $this->l('Save'),
				)
			),
		);
		
		$helper = new HelperForm();
		$helper->bootstrap = true;
		$helper->className = 'bootstrap';
		$helper->show_toolbar = false;
		$helper->table = $this->table;
		
		$lang = new Language((int)Configuration::get('PS_LANG_DEFAULT'));
		$helper->default_form_language = $lang->id;
		$helper->allow_employee_form_lang = Configuration::get('PS_BO_ALLOW_EMPLOYEE_FORM_LANG') ? Configuration::get('PS_BO_ALLOW_EMPLOYEE_FORM_LANG') : 0;
		
		$this->fields_form = array();
		$helper->id = 'firstdata';
		$helper->identifier = $this->identifier;
		$helper->submit_action = 'saveConfig';
		$helper->currentIndex = $this->context->link->getAdminLink('AdminModules', false).'&configure='.$this->name.'&tab_module='.$this->tab.'&module_name='.$this->name;
		$helper->token = $this->token;
		$helper->tpl_vars = array(
			'fields_value' => array(
				'FIRSTDATA_GETEWAYADDR' => Tools::getValue('FIRSTDATA_GETEWAYADDR', Configuration::get('FIRSTDATA_GETEWAYADDR')),
				'FIRSTDATA_POSID' => Tools::getValue('FIRSTDATA_POSID', Configuration::get('FIRSTDATA_POSID')),
				'FIRSTDATA_SHAREDKEY' => Tools::getValue('FIRSTDATA_SHAREDKEY', Configuration::get('FIRSTDATA_SHAREDKEY')),
				'FIRSTDATA_MID' => Tools::getValue('FIRSTDATA_MID', Configuration::get('FIRSTDATA_MID')),
				'FIRSTDATA_USER' => Tools::getValue('FIRSTDATA_USER', Configuration::get('FIRSTDATA_USER')),
				'FIRSTDATA_PASS' => Tools::getValue('FIRSTDATA_PASS', Configuration::get('FIRSTDATA_PASS')),
				'FIRSTDATA_TESTMODE' => Tools::getValue('FIRSTDATA_TESTMODE', Configuration::get('FIRSTDATA_TESTMODE')),
			),
			'languages' => $this->context->controller->getLanguages(),
			'id_language' => $this->context->language->id
		);

		return $helper->generateForm(array($fields_form));
	}
	
	/*
	 * Display transaction list
	 */
	protected function _displayTransactionList()
	{
		# status
		$status = $this->getTransactionStatuses();
		
		
		# redefine params
		$savedParams = $this->context->cookie->fdplist 
			? $this->context->cookie->fdplist 
			: array();
		
		if(Tools::isSubmit('orderBy'))
			$savedParams['orderBy'] = Tools::getValue('orderBy');
		if(Tools::isSubmit('orderWay'))
			$savedParams['orderWay'] = Tools::getValue('orderWay');
		if(Tools::isSubmit('enum'))
			$savedParams['enum'] = Tools::getValue('enum');
		
		# get data
		$savedParams = $this->context->cookie->fdplist;
		
		$orderBy = $savedParams['orderBy'] ? $savedParams['orderBy'] : 'date_add';
		$orderWay = $savedParams['orderWay'] ? $savedParams['orderBy'] : 'DESC';
		$page = Tools::getValue('page', 1);
		$enum = (int) ($savedParams['enum'] ? $savedParams['enum'] : 25);
		
		$data = FirstDataTransaction::getList(
			array(),
			$orderBy, $orderWay,
			$page, $enum
			);
	
		if(!$data['total'])
			return null;
	
		$pagesNum = ceil($data['total'] / $enum);
	
		# assign
		$this->context->smarty->assign(array(
			'baseLink' => $this->context->link->getAdminLink('AdminModules', false).'&configure='.$this->name.'&tab_module='.$this->tab.'&module_name='.$this->name,
		
			'orderBy' => $orderBy,
			'orderWay' => $orderWay,
			'page' => $page,
			'enum' => $enum,
			'pagesNum' => $pagesNum,
			
			'status' => $status,
			
			'data' => $data['records'],
		));
		
		return $this->display(__FILE__, 'trans_list.tpl');
	}

	/*
	 * Hook Payment Option
	 */
    public function hookPaymentOptions($params)
    {
        if (!$this->active)
            return;

        if (!$this->checkCurrency())
            return;

        $newOption = new PrestaShop\PrestaShop\Core\Payment\PaymentOption();
        $newOption->setModuleName($this->name)
                ->setCallToActionText($this->l('Pay online by FirstData'))
                ->setAction($this->context->link->getModuleLink($this->name, 'payment', array(), true))
                ->setLogo( Tools::getShopDomainSsl(true, true).__PS_BASE_URI__.'modules/'.$this->name.'/paymentlogo_min.png' )
                ->setAdditionalInformation($this->display(__FILE__, 'payment_option.tpl'));

        return array(
            $newOption,
		);
    }

	/*
	 * Hook Payment
	 * Called to display firstdata payment anchor
	 */
	public function hookDisplayPayment($params)
	{
		if(!$this->active)
			return;

		$this->smarty->assign(array(
			'this_path' => $this->_path,
			'this_path_bw' => $this->_path,
			'this_path_ssl' => Tools::getShopDomainSsl(true, true) . __PS_BASE_URI__ . 'modules/' . $this->name . '/'
		));
		
		if(version_compare(_PS_VERSION_, '1.7', '<')) {
			$tpl = 'payment_16.tpl';
			$this->context->controller->addCss(_MODULE_DIR_.'firstdatapayment/views/css/fdp.css');
		}
		else $tpl = 'payment_17.tpl';
		
		return $this->display(__FILE__, $tpl);
	}

	public function hookAdminOrder($params) {
		$order = new Order((int)$params['id_order']);
		$status = '--';
		$linked = '';
		if($order->module  == 'firstdatapayment') {
			$id_trans = FirstDataTransaction::getIdByOId($order->id);
			if($id_trans == null) {
				$bro = $order->getBrother();
				foreach($bro as $b) {
					$id_trans = FirstDataTransaction::getIdByOId($b->id);
					if($id_trans) {
						$linked = ' (główne zamówienie powiązane: ' . $b->id . ')';
						break;
					}
				}
			}
			if($id_trans) {
				$trans = new FirstDataTransaction($id_trans);
				$slist = $this->getTransactionStatuses();
				if(isset($slist[$trans->status])) $status = $slist[$trans->status];
				else $status = '???';
			}
		}
		$this->smarty->assign(array(
			'is_fdp' => ($order->module == 'firstdatapayment'),
			'order_id' => $order->id,
			'key' => md5($order->id . Configuration::get('FIRSTDATA_SHAREDKEY')),
			'status' => $status,
			'url' => _MODULE_DIR_ . 'firstdatapayment/ajax.php',
			'fdp_token' => sha1(_COOKIE_KEY_.$this->name),
			'linked' => $linked
		));
		return $this->display(__FILE__, 'adminOrder.tpl');
	}

	public function doAction($action, $id, $key) {
		if(md5($id . Configuration::get('FIRSTDATA_SHAREDKEY')) != $key) return array('error' => $this->l('Security error'));
		switch ($action) {
			case 'R':
				return $this->reload($id);
				break;
			default:
				return array('error' => $this->l('Internal error'));
		}
	}
	
	private function reload($id) {
		$order = new Order($id);
		$oid = $id;
		$id_trans = FirstDataTransaction::getIdByOId($order->id);
		if($id_trans == null) {
			$bro = $order->getBrother();
			foreach($bro as $b) {
				$id_trans = FirstDataTransaction::getIdByOId($b->id);
				if($id_trans) {
					$oid = $b->id;
					break;
				}
			}
		}
		$status_o = 0;
		if($id_trans) {
			$trans = new FirstDataTransaction($id_trans);
			$status_o = $trans->status;
		}
		
		$resp = $this->getTrans($oid, substr($order->date_add, 0, 10));
		if(is_string($resp)) return array('error' => $resp);
		if($resp->orderCode != $oid) return array('error' => 'id error');
		$status_n = $resp->paymentStatus;
		 
		if($status_n != $status_o) {
			$this->updateTrans($order, $status_n, '');
			return array('reload' => 1, 'info' => $this->l('Status has changed, the page will reload'));
		}
		return array('info' => $this->l('Status not changed'));
	}
	
	private function getTrans($id, $d) {
		$api_resp = $this->callApi('/merchant/' . $this->mid . '/payment/' . $id . ';date=' .$d . '?format=json');
		
		if(!is_array($api_resp)) return $api_resp;
		$resp = $api_resp[count($api_resp) - 1];

		return $resp;		
	}

	/*
	 * Hook Payment return
	 */
    public function hookPaymentReturn($params)
    {
        if(!$this->active)
            return;

		$result = Tools::getValue('result');

		/** @var Order $order */
		$order = null;

		if(isset($params['order']))
		{
			$order = $params['order'];
		}
		else
		{
			$orderId = (int) Tools::getValue('id_order');
			$order = new Order($orderId);
		}

        $status = null;

		if($result == 'error')
			$status = 'error';
		elseif($order->current_state == Configuration::get('PS_OS_PAYMENT'))
			$status = 'paid';
		elseif($order->current_state == Configuration::get('FIRSTDATA_ORDERSTATE_PENDING'))
			$status = 'fd_pending';
		elseif($order->current_state == Configuration::get('PS_OS_ERROR'))
			$status = 'error';
        else
			$status = 'pending';

		if(($status == 'error') && (md5($order->id . $this->getSharedKey()) == Tools::getValue('control'))) {
			$customer = new Customer((int)$order->id_customer);
			$cart = new Cart((int)$order->id_cart);
			$currency = new Currency($cart->id_currency);
			//$shop = new Shop($order->id_shop);
			$address = new Address($order->id_address_delivery);
			$country = new Country($address->id_country);

			$orderTotal = (float) $cart->getOrderTotal(true, Cart::BOTH);
			$sessId = md5($order->id . $customer->secure_key);
			$amount = (int) round($orderTotal * 100);

			$data = array(
				'pos_id'		=> $this->posid,
				'order_id'		=> $order->id,
				'session_id'	=> $sessId,
				'amount'		=> $amount,
				'currency'		=> $currency->iso_code,
				'test'			=> $this->testMode ? 'Y' : 'N',
				'language'		=> $this->context->language->iso_code,
				'client_ip'		=> $_SERVER['REMOTE_ADDR'],
				'street'		=> trim( substr($address->address1, 0, 50) ),
				'street_n1'		=> '',
				'street_n2'		=> '',
				'city'			=> trim( $address->city ),
				'postcode'		=> trim( $address->postcode ),
				'country'		=> trim( $country->iso_code ),
				'email'			=> trim( $customer->email ),
				'ba_firstname'	=> trim( $customer->firstname ),
				'ba_lastname'	=> trim( $customer->lastname ),
			);
		
			# to ASCII
			foreach($data as &$field)
				if(!is_numeric($field))
					$field = str_replace("'", '', iconv('UTF-8', 'ascii//TRANSLIT', $field));
		
			$data['controlData'] = $this->_calcControlData($data);
		}
		else $data = array('order_id' => 0);

		$data['status'] = $status;
		$data['contact_url'] = $this->context->link->getPageLink('contact', true);
		$data['url'] = self::GATEWAY_ADDR;
		
		$this->smarty->assign($data);

		return $this->display(__FILE__, 'payment_return.tpl');
    }
	
	/*
	 * Check whether currency is allowed
	 */
	public function checkCurrency()
	{
		$currency = new Currency($this->context->cart->id_currency);
		return in_array($currency->iso_code, $this->avaliableCurrencies);
	}
	
	
	/*
	 * Order validation
	 */
	public function validation()
	{
		# get valid cart
		$cart = $this->context->cart;
		if(!Validate::isLoadedObject($cart) || Order::getOrderByCartId($cart->id))
			return false;

		# validate order
		$currency = new Currency($cart->id_currency);
		$customer = $this->context->customer;
		$orderTotal = (float) $cart->getOrderTotal(true, Cart::BOTH);
		
		$result = $this->validateOrder(
			$cart->id,
			Configuration::get('FIRSTDATA_ORDERSTATE'),
			$orderTotal,
			$this->displayName,
			null,
			array(),
			(int) $currency->id,
			false,
			$customer->secure_key
		);
		if(!$result)
			return false;

		$this->context->cookie->firstdatapaymentOrderId = $this->currentOrder;
		$this->context->cookie->write();

		# get order
		$order = new Order($this->currentOrder);
			
		# make FD request
		$sessId = md5($order->id.$customer->secure_key);
		$amount = (int) round($orderTotal*100);
		$address = new Address($order->id_address_delivery);
		$country = new Country($address->id_country);
		
		$data = array(
			'pos_id'		=> $this->posid,
			'order_id'		=> $order->id,
			'session_id'	=> $sessId,
			'amount'		=> $amount,
			'currency'		=> $currency->iso_code,
			'test'			=> $this->testMode ? 'Y' : 'N',
			'language'		=> $this->context->language->iso_code,
			'client_ip'		=> $_SERVER['REMOTE_ADDR'],
			
			# address
			'street'		=> trim( substr($address->address1, 0,50) ),
			'street_n1'		=> '',
			'street_n2'		=> '',
			'city'			=> trim( $address->city ),
			'postcode'		=> trim( $address->postcode ),
			'country'		=> trim( $country->iso_code ),
			'email'			=> trim( $customer->email ),
			'ba_firstname'	=> trim( $customer->firstname ),
			'ba_lastname'	=> trim( $customer->lastname ),
		);
		
		# to ASCII
		foreach($data as &$field)
			if(!is_numeric($field))
				$field = str_replace("'", '', iconv('UTF-8', 'ascii//TRANSLIT', $field));
		
		$data['controlData'] = $this->_calcControlData($data);

		
		# build link
		$paymentLink = self::GATEWAY_ADDR .'?'.http_build_query($data);
		Tools::redirect($paymentLink);

		return true;
	}
	
	/*
	 * FirstData request
	 */
	public function request()
	{
		if(!$_POST)
		{
			$this->_log('no input');
			return false;
		}
	
		# collect data
		$status = 0;
		$error = false;
		
		if(isset($_POST['err_code']))
		{
			$status = 0;
			$error = $_POST['err_code'];
			$data = array(
				'pos_id'			=> (int) $_POST['pos_id'],
				'order_id'			=> (int) $_POST['order_id'],
				'session_id'		=> $_POST['session_id'],
				'amount'			=> (int) $_POST['amount'],
				'message'			=> $_POST['message'],
				'err_code'			=> $_POST['err_code'],
			);
			if(isset($_POST['transaction_id']))
				$data['transaction_id'] = $_POST['transaction_id'];
		}
		else
		{
			$status = (int) $_POST['response_code'];
			$data = array(
				'pos_id'			=> (int) $_POST['pos_id'],
				'order_id'			=> (int) $_POST['order_id'],
				'session_id'		=> $_POST['session_id'],
				'amount'			=> (int) $_POST['amount'],
				'response_code'		=> (int) $_POST['response_code'],
				'transaction_id'	=> $_POST['transaction_id'],
				'cc_number_hash'	=> isset($_POST['cc_number_hash']) ? $_POST['cc_number_hash'] : null,
				'bin'				=> isset($_POST['bin']) ? $_POST['bin'] : null,
				'card_type'			=> isset($_POST['card_type']) ? $_POST['card_type'] : null,
				'auth_code'			=> isset($_POST['auth_code']) ? $_POST['auth_code'] : null,
			);
		}
		
		# check control data
		$controlData = $_POST['controlData'];
		$calcControlData = $this->_calcControlData($data);
		
		if($controlData != $calcControlData)
		{
			$text = "Wrong controlData in confirmation request.\r\n  order id: ".$data['order_id'];
			if(isset($data['transaction_id']))
				$text .= "\r\n  FD tid: ".$data['transaction_id'];
			$this->_log($text);
			
			return false;
		}
			
		# convert values
		$data['amount'] = $data['amount'] / 100.0;
			
		# load order
		$order = new Order($data['order_id']);
		$customer = new Customer($order->id_customer);
		$sessId = md5($order->id.$customer->secure_key);
		$bro = $order->getBrother();
		
		# check pos id, session id and amount
		if($data['pos_id'] != $this->posid || $data['session_id'] != $sessId)
		{
			# state - error
			$state = (int) Configuration::get('PS_OS_ERROR');
			$this->updOrderStatus($order, $bro, $state);
			
			# log
			$text = "Unauthorized data.\r\n  order id: ".$data['order_id'];
			if(isset($data['transaction_id']))
				$text .= "\r\n  FD tid: ".$data['transaction_id'];
			$this->_log($text);
			
			return false;
		}
		
		$total = $order->total_paid;
		foreach($bro as $b) $total += $b->total_paid;
		if($data['amount'] != $total)
		{
			# state - error
			$state = (int) Configuration::get('PS_OS_ERROR');
			$this->updOrderStatus($order, $bro, $state);
			
			# log
			$text = "Wrong payment amount.\r\n  order id: ".$data['order_id'];
			if(isset($data['transaction_id']))
				$text .= "\r\n  FD tid: ".$data['transaction_id'];
			$this->_log($text);
			
			return false;
		}
		
		# if error
		if($error)
		{
			# update order
			$state = ($error == '005')
				? (int) Configuration::get('PS_OS_CANCELED')
				: (int) Configuration::get('PS_OS_ERROR');
			$this->updOrderStatus($order, $bro, $state);
				
			# error log
			$text = "Payment error.\r\n  error code: ".$data['err_code']."\r\n  msg: ".$data['message']."\r\n  order id: ".$data['order_id'];
			if(isset($data['transaction_id']))
				$text .= "\r\n  FD tid: ".$data['transaction_id'];
			$this->_log($text);
			
			return false;
		}
		
		if(!$this->updateTrans($order, $status, $data['transaction_id'])) return false;

		$text = "Transaction complete\r\n  FD tid: ". $data['transaction_id'];
		$this->_log($text);
		
		return true;
	}

	private function updOrderStatus($order, $bro, $state) {
		if($order->current_state != $state) {
			$order->setCurrentState($state);
			foreach($bro as $b) $b->setCurrentState($state);
		}
	}

	private function updateTrans($order, $status, $transaction_id) {
		$bro = $order->getBrother();
		# load transaction
		$id_trans = FirstDataTransaction::getIdByOId($order->id);
	
		if($id_trans)
		{
			$trans = new FirstDataTransaction($id_trans);
		}
		else
		{
			$trans = new FirstDataTransaction();
			$trans->id_order = $order->id;
			$trans->status = 0;
			$trans->error = 0;
			$trans->fd_tid = $transaction_id;
		
			if(!$trans->add())
			{
				# state - error
				$state = (int) Configuration::get('PS_OS_ERROR');
				$this->updOrderStatus($order, $bro, $state);
		
				# log
				$text = "Can not create transaction in DB.\r\n  order id: " . $order->id;
				if(isset($transaction_id))
					$text .= "\r\n  FD tid: ".$transaction_id;
				$this->_log($text);
							
				return false;
			}
		}
		
		if(in_array($status, array(30, 35)))
		{
			# pay (only one time)
			$paid = $trans->getPaymentId();
			if(!$paid)
			{
				//# payment
				//$order->addOrderPayment($data['amount'], $this->displayName, $transaction_id);
			}
				
			# update trans
			$trans->status = $status;
			$trans->date_accept = date('Y-m-d H:i:s');
			$trans->error = null;
			
			# update order
			$state = (int) Configuration::get('PS_OS_PAYMENT');
			$this->updOrderStatus($order, $bro, $state);
		}
		elseif(in_array($status, array(20)))
		{
			$trans->status = $status;
				
			# update order
			$state = $this->getPendingOs();
			$this->updOrderStatus($order, $bro, $state);
		}
		elseif(in_array($status, array(21, 40)))
		{
			$trans->status = $status;
			$trans->date_cancel = date('Y-m-d H:i:s');
		
			# update order
			$state = (int) Configuration::get('PS_OS_ERROR');
			$this->updOrderStatus($order, $bro, $state);
		}
		else
		{
			$trans->status = $status;
		}
		
		# save transaction
		$result = $trans->save();
		if(!$result)
		{
			# state - error
			$state = (int) Configuration::get('PS_OS_ERROR');
			$this->updOrderStatus($order, $bro, $state);
					
			# log
			$text = "Can not update transaction in DB.\r\n  order id: " . $order->id;
			if(isset($transaction_id))
				$text .= "\r\n  FD tid: ".$transaction_id;
			$this->_log($text);
						
			return false;
		}
		return true;
	}

	private function getPendingOs() {
		$id = Configuration::get('FIRSTDATA_ORDERSTATE_PENDING');
		if(!$id) {
			$state2 = new OrderState();
			$state2->module_name = $this->name;
			$state2->name = array();
			$this->_languages = Language::getLanguages(false);
			foreach($this->_languages as $lang)
				$state2->name[(int)$lang['id_lang']] = $this->l('Payment by FirstData - pending');
			$state2->color = '#4169E1';
			$state2->add();
			$source = dirname(__FILE__).'/logo.gif';
			$destination = dirname(__FILE__).'/../../img/os/'.(int)$state2->id.'.gif';
			copy($source, $destination);
			Configuration::updateValue('FIRSTDATA_ORDERSTATE_PENDING', $state2->id);					
		}
		return (int) $id;
	}

	/*
	 * Calc control data for FD request
	 */
	protected function _calcControlData($data)
	{
		# filter empty values
		if(!function_exists('_calcControlData_sub'))
		{
			function _calcControlData_sub($value) { return (bool) $value; }
		}
		
		$data = array_filter($data, '_calcControlData_sub');
		# build query
		$query = '';
		foreach($data as $key => $value)
			$query .= ($query ? '&' : '') . $key .'='. $value;
		
		# add shared key binary string
		$l = strlen($this->sharedKey)/2;
		$sharedKeyBin = '';
		for($i=1; $i<=$l; ++$i)
			$sharedKeyBin .= pack('H*', substr($this->sharedKey, 2*$i-2, 2));
		
		$query .= $sharedKeyBin;
		
		# hash
		return hash('sha256', $query);
	}
		
	protected function _log($text)
	{
		$text = date('Y-m-d H:i:s') .' '. $text ."\r\n\r\n";
		
		$fp = fopen($this->logFile, 'a');
		fwrite($fp, $text);
		fclose($fp);
	}
	
	
	
}
